#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import logging
import os
import platform
import time
from concurrent.futures.thread import ThreadPoolExecutor
from os import environ, path
from unittest import mock

try:
    from snowflake.connector.util_text import random_string
except ImportError:
    from ..randomize import random_string

import pytest

import snowflake.connector.ocsp_snowflake
from snowflake.connector import OperationalError
from snowflake.connector.errors import RevocationCheckError
from snowflake.connector.ocsp_asn1crypto import SnowflakeOCSPAsn1Crypto as SFOCSP
from snowflake.connector.ocsp_snowflake import OCSPCache, OCSPServer, SnowflakeOCSP
from snowflake.connector.ssl_wrap_socket import _openssl_connect

try:
    from snowflake.connector.cache import SFDictFileCache
    from snowflake.connector.errorcode import (
        ER_OCSP_RESPONSE_CERT_STATUS_REVOKED,
        ER_OCSP_RESPONSE_FETCH_FAILURE,
    )
    from snowflake.connector.ocsp_snowflake import OCSP_CACHE

    @pytest.fixture(autouse=True)
    def overwrite_ocsp_cache(tmpdir):
        """This fixture swaps out the actual OCSP cache for a temprary one."""
        if OCSP_CACHE is not None:
            tmp_cache_file = os.path.join(tmpdir, "tmp_cache")
            with mock.patch(
                "snowflake.connector.ocsp_snowflake.OCSP_CACHE",
                SFDictFileCache(file_path=tmp_cache_file),
            ):
                yield
            os.unlink(tmp_cache_file)

except ImportError:
    ER_OCSP_RESPONSE_CERT_STATUS_REVOKED = None
    ER_OCSP_RESPONSE_FETCH_FAILURE = None
    OCSP_CACHE = None

TARGET_HOSTS = [
    "ocspssd.us-east-1.snowflakecomputing.com",
    "sqs.us-west-2.amazonaws.com",
    "sfcsupport.us-east-1.snowflakecomputing.com",
    "sfcsupport.eu-central-1.snowflakecomputing.com",
    "sfc-eng-regression.s3.amazonaws.com",
    "sfctest0.snowflakecomputing.com",
    "sfc-ds2-customer-stage.s3.amazonaws.com",
    "snowflake.okta.com",
    "sfcdev1.blob.core.windows.net",
    "sfc-aus-ds1-customer-stage.s3-ap-southeast-2.amazonaws.com",
]

THIS_DIR = path.dirname(path.realpath(__file__))


@pytest.fixture(autouse=True)
def random_ocsp_response_validation_cache():
    file_path = {
        "linux": os.path.join(
            "~",
            ".cache",
            "snowflake",
            f"ocsp_response_validation_cache{random_string()}",
        ),
        "darwin": os.path.join(
            "~",
            "Library",
            "Caches",
            "Snowflake",
            f"ocsp_response_validation_cache{random_string()}",
        ),
        "windows": os.path.join(
            "~",
            "AppData",
            "Local",
            "Snowflake",
            "Caches",
            f"ocsp_response_validation_cache{random_string()}",
        ),
    }
    yield SFDictFileCache(
        entry_lifetime=3600,
        file_path=file_path,
    )
    try:
        os.unlink(file_path[platform.system().lower()])
    except Exception:
        pass


def test_ocsp():
    """OCSP tests."""
    # reset the memory cache
    SnowflakeOCSP.clear_cache()
    ocsp = SFOCSP()
    for url in TARGET_HOSTS:
        connection = _openssl_connect(url, timeout=5)
        assert ocsp.validate(url, connection), f"Failed to validate: {url}"


def test_ocsp_wo_cache_server():
    """OCSP Tests with Cache Server Disabled."""
    SnowflakeOCSP.clear_cache()
    ocsp = SFOCSP(use_ocsp_cache_server=False)
    for url in TARGET_HOSTS:
        connection = _openssl_connect(url)
        assert ocsp.validate(url, connection), f"Failed to validate: {url}"


def test_ocsp_wo_cache_file():
    """OCSP tests without File cache.

    Notes:
        Use /etc as a readonly directory such that no cache file is used.
    """
    # reset the memory cache
    SnowflakeOCSP.clear_cache()
    OCSPCache.del_cache_file()
    environ["SF_OCSP_RESPONSE_CACHE_DIR"] = "/etc"
    OCSPCache.reset_cache_dir()

    try:
        ocsp = SFOCSP()
        for url in TARGET_HOSTS:
            connection = _openssl_connect(url)
            assert ocsp.validate(url, connection), f"Failed to validate: {url}"
    finally:
        del environ["SF_OCSP_RESPONSE_CACHE_DIR"]
        OCSPCache.reset_cache_dir()


def test_ocsp_fail_open_w_single_endpoint():
    SnowflakeOCSP.clear_cache()

    OCSPCache.del_cache_file()

    environ["SF_OCSP_TEST_MODE"] = "true"
    environ["SF_TEST_OCSP_URL"] = "http://httpbin.org/delay/10"
    environ["SF_TEST_CA_OCSP_RESPONDER_CONNECTION_TIMEOUT"] = "5"

    ocsp = SFOCSP(use_ocsp_cache_server=False)
    connection = _openssl_connect("snowflake.okta.com")

    try:
        assert ocsp.validate(
            "snowflake.okta.com", connection
        ), "Failed to validate: {}".format("snowflake.okta.com")
    finally:
        del environ["SF_OCSP_TEST_MODE"]
        del environ["SF_TEST_OCSP_URL"]
        del environ["SF_TEST_CA_OCSP_RESPONDER_CONNECTION_TIMEOUT"]


@pytest.mark.skipif(
    ER_OCSP_RESPONSE_CERT_STATUS_REVOKED is None,
    reason="No ER_OCSP_RESPONSE_CERT_STATUS_REVOKED is available.",
)
def test_ocsp_fail_close_w_single_endpoint():
    SnowflakeOCSP.clear_cache()

    environ["SF_OCSP_TEST_MODE"] = "true"
    environ["SF_TEST_OCSP_URL"] = "http://httpbin.org/delay/10"
    environ["SF_TEST_CA_OCSP_RESPONDER_CONNECTION_TIMEOUT"] = "5"

    OCSPCache.del_cache_file()

    ocsp = SFOCSP(use_ocsp_cache_server=False, use_fail_open=False)
    connection = _openssl_connect("snowflake.okta.com")

    with pytest.raises(RevocationCheckError) as ex:
        ocsp.validate("snowflake.okta.com", connection)

    try:
        assert (
            ex.value.errno == ER_OCSP_RESPONSE_FETCH_FAILURE
        ), "Connection should have failed"
    finally:
        del environ["SF_OCSP_TEST_MODE"]
        del environ["SF_TEST_OCSP_URL"]
        del environ["SF_TEST_CA_OCSP_RESPONDER_CONNECTION_TIMEOUT"]


def test_ocsp_bad_validity():
    SnowflakeOCSP.clear_cache()

    environ["SF_OCSP_TEST_MODE"] = "true"
    environ["SF_TEST_OCSP_FORCE_BAD_RESPONSE_VALIDITY"] = "true"

    OCSPCache.del_cache_file()

    ocsp = SFOCSP(use_ocsp_cache_server=False)
    connection = _openssl_connect("snowflake.okta.com")

    assert ocsp.validate(
        "snowflake.okta.com", connection
    ), "Connection should have passed with fail open"
    del environ["SF_OCSP_TEST_MODE"]
    del environ["SF_TEST_OCSP_FORCE_BAD_RESPONSE_VALIDITY"]


def test_ocsp_single_endpoint():
    environ["SF_OCSP_ACTIVATE_NEW_ENDPOINT"] = "True"
    SnowflakeOCSP.clear_cache()
    ocsp = SFOCSP()
    ocsp.OCSP_CACHE_SERVER.NEW_DEFAULT_CACHE_SERVER_BASE_URL = "https://snowflake.preprod3.us-west-2-dev.external-zone.snowflakecomputing.com:8085/ocsp/"
    connection = _openssl_connect("snowflake.okta.com")
    assert ocsp.validate(
        "snowflake.okta.com", connection
    ), "Failed to validate: {}".format("snowflake.okta.com")

    del environ["SF_OCSP_ACTIVATE_NEW_ENDPOINT"]


def test_ocsp_by_post_method():
    """OCSP tests."""
    # reset the memory cache
    SnowflakeOCSP.clear_cache()
    ocsp = SFOCSP(use_post_method=True)
    for url in TARGET_HOSTS:
        connection = _openssl_connect(url)
        assert ocsp.validate(url, connection), f"Failed to validate: {url}"


def test_ocsp_with_file_cache(tmpdir):
    """OCSP tests and the cache server and file."""
    tmp_dir = str(tmpdir.mkdir("ocsp_response_cache"))
    cache_file_name = path.join(tmp_dir, "cache_file.txt")

    # reset the memory cache
    SnowflakeOCSP.clear_cache()
    ocsp = SFOCSP(ocsp_response_cache_uri="file://" + cache_file_name)
    for url in TARGET_HOSTS:
        connection = _openssl_connect(url)
        assert ocsp.validate(url, connection), f"Failed to validate: {url}"


@pytest.mark.skipolddriver
def test_ocsp_with_bogus_cache_files(tmpdir, random_ocsp_response_validation_cache):
    with mock.patch(
        "snowflake.connector.ocsp_snowflake.OCSP_RESPONSE_VALIDATION_CACHE",
        random_ocsp_response_validation_cache,
    ):
        from snowflake.connector.ocsp_snowflake import OCSPResponseValidationResult

        """Attempts to use bogus OCSP response data."""
        cache_file_name, target_hosts = _store_cache_in_file(tmpdir)

        ocsp = SFOCSP()
        OCSPCache.read_ocsp_response_cache_file(ocsp, cache_file_name)
        cache_data = snowflake.connector.ocsp_snowflake.OCSP_RESPONSE_VALIDATION_CACHE
        assert cache_data, "more than one cache entries should be stored."

        # setting bogus data
        current_time = int(time.time())
        for k, _ in cache_data.items():
            cache_data[k] = OCSPResponseValidationResult(
                ocsp_response=b"bogus",
                ts=current_time,
                validated=True,
            )

        # write back the cache file
        OCSPCache.CACHE = cache_data
        OCSPCache.write_ocsp_response_cache_file(ocsp, cache_file_name)

        # forces to use the bogus cache file but it should raise errors
        SnowflakeOCSP.clear_cache()
        ocsp = SFOCSP()
        for hostname in target_hosts:
            connection = _openssl_connect(hostname)
            assert ocsp.validate(hostname, connection), "Failed to validate: {}".format(
                hostname
            )


@pytest.mark.skipolddriver
def test_ocsp_with_outdated_cache(tmpdir, random_ocsp_response_validation_cache):
    with mock.patch(
        "snowflake.connector.ocsp_snowflake.OCSP_RESPONSE_VALIDATION_CACHE",
        random_ocsp_response_validation_cache,
    ):
        from snowflake.connector.ocsp_snowflake import OCSPResponseValidationResult

        """Attempts to use outdated OCSP response cache file."""
        cache_file_name, target_hosts = _store_cache_in_file(tmpdir)

        ocsp = SFOCSP()

        # reading cache file
        OCSPCache.read_ocsp_response_cache_file(ocsp, cache_file_name)
        cache_data = snowflake.connector.ocsp_snowflake.OCSP_RESPONSE_VALIDATION_CACHE
        assert cache_data, "more than one cache entries should be stored."

        # setting outdated data
        current_time = int(time.time())
        for k, v in cache_data.items():
            cache_data[k] = OCSPResponseValidationResult(
                ocsp_response=v.ocsp_response,
                ts=current_time - 144 * 60 * 60,
                validated=True,
            )

        # write back the cache file
        OCSPCache.CACHE = cache_data
        OCSPCache.write_ocsp_response_cache_file(ocsp, cache_file_name)

        # forces to use the bogus cache file but it should raise errors
        SnowflakeOCSP.clear_cache()  # reset the memory cache
        SFOCSP()
        assert (
            SnowflakeOCSP.cache_size() == 0
        ), "must be empty. outdated cache should not be loaded"


def _store_cache_in_file(tmpdir, target_hosts=None):
    if target_hosts is None:
        target_hosts = TARGET_HOSTS
    os.environ["SF_OCSP_RESPONSE_CACHE_DIR"] = str(tmpdir)
    OCSPCache.reset_cache_dir()
    filename = path.join(str(tmpdir), "ocsp_response_cache.json")

    # cache OCSP response
    SnowflakeOCSP.clear_cache()
    ocsp = SFOCSP(
        ocsp_response_cache_uri="file://" + filename, use_ocsp_cache_server=False
    )
    for hostname in target_hosts:
        connection = _openssl_connect(hostname)
        assert ocsp.validate(hostname, connection), "Failed to validate: {}".format(
            hostname
        )
    assert path.exists(filename), "OCSP response cache file"
    return filename, target_hosts


def test_ocsp_with_invalid_cache_file():
    """OCSP tests with an invalid cache file."""
    SnowflakeOCSP.clear_cache()  # reset the memory cache
    ocsp = SFOCSP(ocsp_response_cache_uri="NEVER_EXISTS")
    for url in TARGET_HOSTS[0:1]:
        connection = _openssl_connect(url)
        assert ocsp.validate(url, connection), f"Failed to validate: {url}"


@mock.patch(
    "snowflake.connector.ocsp_snowflake.SnowflakeOCSP._fetch_ocsp_response",
    side_effect=BrokenPipeError("fake error"),
)
def test_ocsp_cache_when_server_is_down(
    mock_fetch_ocsp_response, tmpdir, random_ocsp_response_validation_cache
):
    with mock.patch(
        "snowflake.connector.ocsp_snowflake.OCSP_RESPONSE_VALIDATION_CACHE",
        random_ocsp_response_validation_cache,
    ):
        ocsp = SFOCSP()

        """Attempts to use outdated OCSP response cache file."""
        cache_file_name, target_hosts = _store_cache_in_file(tmpdir)

        # reading cache file
        OCSPCache.read_ocsp_response_cache_file(ocsp, cache_file_name)
        cache_data = snowflake.connector.ocsp_snowflake.OCSP_RESPONSE_VALIDATION_CACHE
        assert not cache_data, "no cache should present because of broken pipe"


def test_concurrent_ocsp_requests(tmpdir):
    """Run OCSP revocation checks in parallel. The memory and file caches are deleted randomly."""
    cache_file_name = path.join(str(tmpdir), "cache_file.txt")
    SnowflakeOCSP.clear_cache()  # reset the memory cache

    target_hosts = TARGET_HOSTS * 5
    pool = ThreadPoolExecutor(len(target_hosts))
    for hostname in target_hosts:
        pool.submit(_validate_certs_using_ocsp, hostname, cache_file_name)
    pool.shutdown()


def _validate_certs_using_ocsp(url, cache_file_name):
    """Validate OCSP response. Deleting memory cache and file cache randomly."""
    logger = logging.getLogger("test")
    import random
    import time

    time.sleep(random.randint(0, 3))
    if random.random() < 0.2:
        logger.info("clearing up cache: OCSP_VALIDATION_CACHE")
        SnowflakeOCSP.clear_cache()
    if random.random() < 0.05:
        logger.info("deleting a cache file: %s", cache_file_name)
        SnowflakeOCSP.delete_cache_file()

    connection = _openssl_connect(url)
    ocsp = SFOCSP(ocsp_response_cache_uri="file://" + cache_file_name)
    ocsp.validate(url, connection)


@pytest.mark.skip(reason="certificate expired.")
def test_ocsp_revoked_certificate():
    """Tests revoked certificate."""
    revoked_cert = path.join(THIS_DIR, "../data", "cert_tests", "revoked_certs.pem")

    SnowflakeOCSP.clear_cache()  # reset the memory cache
    ocsp = SFOCSP()

    with pytest.raises(OperationalError) as ex:
        ocsp.validate_certfile(revoked_cert)
    assert ex.value.errno == ex.value.errno == ER_OCSP_RESPONSE_CERT_STATUS_REVOKED


def test_ocsp_incomplete_chain():
    """Tests incomplete chained certificate."""
    incomplete_chain_cert = path.join(
        THIS_DIR, "../data", "cert_tests", "incomplete-chain.pem"
    )

    SnowflakeOCSP.clear_cache()  # reset the memory cache
    ocsp = SFOCSP()

    with pytest.raises(OperationalError) as ex:
        ocsp.validate_certfile(incomplete_chain_cert)
    assert "CA certificate is NOT found" in ex.value.msg


def test_building_retry_url():
    # privatelink retry url
    OCSP_SERVER = OCSPServer()
    OCSP_SERVER.OCSP_RETRY_URL = None
    OCSP_SERVER.CACHE_SERVER_URL = (
        "http://ocsp.us-east-1.snowflakecomputing.com/ocsp_response_cache.json"
    )
    OCSP_SERVER.reset_ocsp_dynamic_cache_server_url(None)
    assert (
        OCSP_SERVER.OCSP_RETRY_URL
        == "http://ocsp.us-east-1.snowflakecomputing.com/retry/{0}/{1}"
    )

    assert (
        OCSP_SERVER.generate_get_url("http://oneocsp.microsoft.com", "1234")
        == "http://ocsp.us-east-1.snowflakecomputing.com/retry/oneocsp.microsoft.com/1234"
    )
    assert (
        OCSP_SERVER.generate_get_url("http://oneocsp.microsoft.com/", "1234")
        == "http://ocsp.us-east-1.snowflakecomputing.com/retry/oneocsp.microsoft.com/1234"
    )
    assert (
        OCSP_SERVER.generate_get_url("http://oneocsp.microsoft.com/ocsp", "1234")
        == "http://ocsp.us-east-1.snowflakecomputing.com/retry/oneocsp.microsoft.com/ocsp/1234"
    )

    # ensure we also handle port
    assert (
        OCSP_SERVER.generate_get_url("http://oneocsp.microsoft.com:8080", "1234")
        == "http://ocsp.us-east-1.snowflakecomputing.com/retry/oneocsp.microsoft.com:8080/1234"
    )
    assert (
        OCSP_SERVER.generate_get_url("http://oneocsp.microsoft.com:8080/", "1234")
        == "http://ocsp.us-east-1.snowflakecomputing.com/retry/oneocsp.microsoft.com:8080/1234"
    )
    assert (
        OCSP_SERVER.generate_get_url("http://oneocsp.microsoft.com:8080/ocsp", "1234")
        == "http://ocsp.us-east-1.snowflakecomputing.com/retry/oneocsp.microsoft.com:8080/ocsp/1234"
    )

    # privatelink retry url with port
    OCSP_SERVER.OCSP_RETRY_URL = None
    OCSP_SERVER.CACHE_SERVER_URL = (
        "http://ocsp.us-east-1.snowflakecomputing.com:80/ocsp_response_cache" ".json"
    )
    OCSP_SERVER.reset_ocsp_dynamic_cache_server_url(None)
    assert (
        OCSP_SERVER.OCSP_RETRY_URL
        == "http://ocsp.us-east-1.snowflakecomputing.com:80/retry/{0}/{1}"
    )

    # non-privatelink retry url
    OCSP_SERVER.OCSP_RETRY_URL = None
    OCSP_SERVER.CACHE_SERVER_URL = (
        "http://ocsp.snowflakecomputing.com/ocsp_response_cache.json"
    )
    OCSP_SERVER.reset_ocsp_dynamic_cache_server_url(None)
    assert OCSP_SERVER.OCSP_RETRY_URL is None

    # non-privatelink retry url with port
    OCSP_SERVER.OCSP_RETRY_URL = None
    OCSP_SERVER.CACHE_SERVER_URL = (
        "http://ocsp.snowflakecomputing.com:80/ocsp_response_cache.json"
    )
    OCSP_SERVER.reset_ocsp_dynamic_cache_server_url(None)
    assert OCSP_SERVER.OCSP_RETRY_URL is None


def test_building_new_retry():
    OCSP_SERVER = OCSPServer()
    OCSP_SERVER.OCSP_RETRY_URL = None
    hname = "a1.us-east-1.snowflakecomputing.com"
    os.environ["SF_OCSP_ACTIVATE_NEW_ENDPOINT"] = "true"
    OCSP_SERVER.reset_ocsp_endpoint(hname)
    assert (
        OCSP_SERVER.CACHE_SERVER_URL
        == "https://ocspssd.us-east-1.snowflakecomputing.com/ocsp/fetch"
    )

    assert (
        OCSP_SERVER.OCSP_RETRY_URL
        == "https://ocspssd.us-east-1.snowflakecomputing.com/ocsp/retry"
    )

    hname = "a1-12345.global.snowflakecomputing.com"
    OCSP_SERVER.reset_ocsp_endpoint(hname)
    assert (
        OCSP_SERVER.CACHE_SERVER_URL
        == "https://ocspssd-12345.global.snowflakecomputing.com/ocsp/fetch"
    )

    assert (
        OCSP_SERVER.OCSP_RETRY_URL
        == "https://ocspssd-12345.global.snowflakecomputing.com/ocsp/retry"
    )

    hname = "snowflake.okta.com"
    OCSP_SERVER.reset_ocsp_endpoint(hname)
    assert (
        OCSP_SERVER.CACHE_SERVER_URL
        == "https://ocspssd.snowflakecomputing.com/ocsp/fetch"
    )

    assert (
        OCSP_SERVER.OCSP_RETRY_URL
        == "https://ocspssd.snowflakecomputing.com/ocsp/retry"
    )

    del os.environ["SF_OCSP_ACTIVATE_NEW_ENDPOINT"]



#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

"""Various constants."""

from __future__ import annotations

import platform
import sys

from .version import VERSION

SNOWFLAKE_CONNECTOR_VERSION = ".".join(str(v) for v in VERSION[0:3])
PYTHON_VERSION = ".".join(str(v) for v in sys.version_info[:3])
OPERATING_SYSTEM = platform.system()
PLATFORM = platform.platform()
IMPLEMENTATION = platform.python_implementation()
COMPILER = platform.python_compiler()

CLIENT_NAME = "PythonConnector"  # don't change!
CLIENT_VERSION = ".".join([str(v) for v in VERSION[:3]])



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

try:
    from snowflake.connector.url_util import is_valid_url, url_encode_str
except ImportError:

    def is_valid_url(s):
        return False

    def url_encode_str(s):
        return ""


def test_url_validator():
    assert is_valid_url("https://ssoTestURL.okta.com")
    assert is_valid_url("https://ssoTestURL.okta.com:8080")
    assert is_valid_url("https://ssoTestURL.okta.com/testpathvalue")

    assert not is_valid_url("-a Calculator")
    assert not is_valid_url("This is a random text")
    assert not is_valid_url("file://TestForFile")


def test_encoder():
    assert url_encode_str("Hello @World") == "Hello+%40World"
    assert url_encode_str("Test//String") == "Test%2F%2FString"
    assert url_encode_str(None) == ""



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import pytest

from snowflake.connector.telemetry_oob import TelemetryService


@pytest.fixture(autouse=True, scope="session")
def disable_oob_telemetry():
    oob_telemetry_service = TelemetryService.get_instance()
    original_state = oob_telemetry_service.enabled
    oob_telemetry_service.disable()
    yield None
    if original_state:
        oob_telemetry_service.enable()



#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from ..errorcode import ER_NO_PASSWORD
from ..errors import ProgrammingError
from .by_plugin import AuthByPlugin, AuthType

if TYPE_CHECKING:
    from .. import SnowflakeConnection

logger = logging.getLogger(__name__)

MFA_TOKEN = "MFATOKEN"


class AuthByUsrPwdMfa(AuthByPlugin):
    """Username & password & mfa authenticator."""

    @property
    def assertion_content(self) -> str:
        return "*********"

    def __init__(
        self,
        password: str,
        mfa_token: str | None = None,
        **kwargs,
    ) -> None:
        """Initializes and instance with a password and a mfa token."""
        super().__init__(**kwargs)
        self._password: str | None = password
        self._mfa_token: str | None = mfa_token

    def reset_secrets(self) -> None:
        self._password = None
        self._mfa_token = None

    @property
    def type_(self) -> AuthType:
        return AuthType.USR_PWD_MFA

    def prepare(
        self,
        *,
        conn: SnowflakeConnection,
        **kwargs: Any,
    ) -> None:
        if conn._rest and conn._rest.mfa_token:
            self._mfa_token = conn._rest.mfa_token

    def reauthenticate(self, **kwargs) -> dict[str, bool]:
        return {"success": False}

    def update_body(self, body: dict[Any, Any]) -> None:
        """Sets the password and mfa_token if available.

        Don't set body['data']['AUTHENTICATOR'], since this is still snowflake default authenticator.
        """
        if not self._password:
            raise ProgrammingError(
                msg="Password for username password authenticator is empty.",
                errno=ER_NO_PASSWORD,
            )
        body["data"]["PASSWORD"] = self._password
        if self._mfa_token:
            body["data"]["TOKEN"] = self._mfa_token



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import os
import shutil
import tempfile
import threading
import time
from abc import ABC, abstractmethod
from collections import defaultdict
from io import BytesIO
from logging import getLogger
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, NamedTuple

import OpenSSL

from .constants import HTTP_HEADER_CONTENT_ENCODING, FileHeader, ResultStatus
from .encryption_util import EncryptionMetadata, SnowflakeEncryptionUtil
from .errors import RequestExceedMaxRetryError
from .file_util import SnowflakeFileUtil
from .vendored import requests
from .vendored.requests import ConnectionError, Timeout
from .vendored.urllib3 import HTTPResponse

if TYPE_CHECKING:  # pragma: no cover
    from .file_transfer_agent import SnowflakeFileMeta, StorageCredential

logger = getLogger(__name__)


class SnowflakeFileEncryptionMaterial(NamedTuple):
    query_stage_master_key: str  # query stage master key
    query_id: str  # query id
    smk_id: int  # SMK id


METHODS = {
    "GET": requests.get,
    "PUT": requests.put,
    "POST": requests.post,
    "HEAD": requests.head,
    "DELETE": requests.delete,
}


def remove_content_encoding(resp: requests.Response, **kwargs) -> None:
    """Remove content-encoding header and decoder so decompression is not triggered"""
    if HTTP_HEADER_CONTENT_ENCODING in resp.headers:
        if isinstance(resp.raw, HTTPResponse):
            resp.raw._decoder = None
            resp.raw.headers.pop(HTTP_HEADER_CONTENT_ENCODING)


class SnowflakeStorageClient(ABC):
    TRANSIENT_HTTP_ERR = (408, 429, 500, 502, 503, 504)

    TRANSIENT_ERRORS = (OpenSSL.SSL.SysCallError, Timeout, ConnectionError)
    SLEEP_MAX = 16.0
    SLEEP_UNIT = 1.0

    def __init__(
        self,
        meta: SnowflakeFileMeta,
        stage_info: dict[str, Any],
        chunk_size: int,
        chunked_transfer: bool | None = True,
        credentials: StorageCredential | None = None,
        max_retry: int = 5,
    ) -> None:
        self.meta = meta
        self.stage_info = stage_info
        self.retry_count: dict[int | str, int] = defaultdict(lambda: 0)
        self.tmp_dir = tempfile.mkdtemp()
        self.data_file: str | None = None
        self.encryption_metadata: EncryptionMetadata | None = None

        self.max_retry = max_retry  # TODO
        self.credentials = credentials
        # UPLOAD
        meta.real_src_file_name = meta.src_file_name
        meta.upload_size = meta.src_file_size
        self.preprocessed = (
            False  # so we don't repeat compression/file digest when re-encrypting
        )
        # DOWNLOAD
        self.full_dst_file_name: str | None = (
            os.path.join(
                self.meta.local_location, os.path.basename(self.meta.dst_file_name)
            )
            if self.meta.local_location
            else None
        )
        self.intermediate_dst_path: Path | None = (
            Path(self.full_dst_file_name + ".part")
            if self.meta.local_location
            else None
        )
        # CHUNK
        self.chunked_transfer = chunked_transfer  # only true for GCS
        self.chunk_size = chunk_size
        self.num_of_chunks = 0
        self.lock = threading.Lock()
        self.successful_transfers: int = 0
        self.failed_transfers: int = 0
        # only used when PRESIGNED_URL expires
        self.last_err_is_presigned_url = False

    def compress(self) -> None:
        if self.meta.require_compress:
            meta = self.meta
            logger.debug(f"compressing file={meta.src_file_name}")
            if meta.intermediate_stream:
                (
                    meta.src_stream,
                    upload_size,
                ) = SnowflakeFileUtil.compress_with_gzip_from_stream(
                    meta.intermediate_stream
                )
            else:
                (
                    meta.real_src_file_name,
                    upload_size,
                ) = SnowflakeFileUtil.compress_file_with_gzip(
                    meta.src_file_name, self.tmp_dir
                )

    def get_digest(self) -> None:
        meta = self.meta
        logger.debug(f"getting digest file={meta.real_src_file_name}")
        if meta.intermediate_stream is None:
            (
                meta.sha256_digest,
                meta.upload_size,
            ) = SnowflakeFileUtil.get_digest_and_size_for_file(meta.real_src_file_name)
        else:
            (
                meta.sha256_digest,
                meta.upload_size,
            ) = SnowflakeFileUtil.get_digest_and_size_for_stream(
                meta.src_stream or meta.intermediate_stream
            )

    def encrypt(self) -> None:
        meta = self.meta
        logger.debug(f"encrypting file={meta.real_src_file_name}")
        if meta.intermediate_stream is None:
            (
                self.encryption_metadata,
                self.data_file,
            ) = SnowflakeEncryptionUtil.encrypt_file(
                meta.encryption_material,
                meta.real_src_file_name,
                tmp_dir=self.tmp_dir,
            )
            meta.upload_size = os.path.getsize(self.data_file)
        else:
            encrypted_stream = BytesIO()
            src_stream = meta.src_stream or meta.intermediate_stream
            src_stream.seek(0)
            self.encryption_metadata = SnowflakeEncryptionUtil.encrypt_stream(
                meta.encryption_material, src_stream, encrypted_stream
            )
            src_stream.seek(0)
            meta.upload_size = encrypted_stream.seek(0, os.SEEK_END)
            encrypted_stream.seek(0)
            if meta.src_stream is not None:
                meta.src_stream.close()
            meta.src_stream = encrypted_stream
            self.data_file = meta.real_src_file_name

    @abstractmethod
    def get_file_header(self, filename: str) -> FileHeader | None:
        """Check if file exists in target location and obtain file metadata if exists.

        Notes:
            Updates meta.result_status.
        """
        pass

    def preprocess(self) -> None:
        meta = self.meta
        logger.debug(f"Preprocessing {meta.src_file_name}")

        file_header = self.get_file_header(
            meta.dst_file_name
        )  # check if file exists on remote
        if not meta.overwrite:
            self.get_digest()  # self.get_file_header needs digest for multiparts upload when aws is used.
            if meta.result_status == ResultStatus.UPLOADED:
                # Skipped
                logger.debug(
                    f'file already exists location="{self.stage_info["location"]}", '
                    f'file_name="{meta.dst_file_name}"'
                )
                meta.dst_file_size = 0
                meta.result_status = ResultStatus.SKIPPED
                self.preprocessed = True
                return
        # Uploading
        if meta.require_compress:
            self.compress()
        self.get_digest()

        if (
            meta.skip_upload_on_content_match
            and file_header
            and meta.sha256_digest == file_header.digest
        ):
            logger.debug(f"same file contents for {meta.name}, skipping upload")
            meta.result_status = ResultStatus.SKIPPED

        self.preprocessed = True

    def prepare_upload(self) -> None:
        meta = self.meta

        if not self.preprocessed:
            self.preprocess()
        elif meta.encryption_material:
            # need to clean up previous encrypted file
            os.remove(self.data_file)

        logger.debug(f"Preparing to upload {meta.src_file_name}")

        if meta.encryption_material:
            self.encrypt()
        else:
            self.data_file = meta.real_src_file_name
        logger.debug("finished preprocessing")
        if meta.upload_size < meta.multipart_threshold or not self.chunked_transfer:
            self.num_of_chunks = 1
        else:
            self.num_of_chunks = ceil(meta.upload_size / self.chunk_size)
        logger.debug(f"number of chunks {self.num_of_chunks}")
        # clean up
        self.retry_count = {}

        for chunk_id in range(self.num_of_chunks):
            self.retry_count[chunk_id] = 0
        if self.chunked_transfer and self.num_of_chunks > 1:
            self._initiate_multipart_upload()

    def finish_upload(self) -> None:
        meta = self.meta
        if self.successful_transfers == self.num_of_chunks:
            if self.num_of_chunks > 1:
                self._complete_multipart_upload()
            meta.result_status = ResultStatus.UPLOADED
            meta.dst_file_size = meta.upload_size
            logger.debug(f"{meta.src_file_name} upload is completed.")
        else:
            # TODO: add more error details to result/meta
            meta.dst_file_size = 0
            logger.debug(f"{meta.src_file_name} upload is aborted.")
            if self.num_of_chunks > 1:
                self._abort_multipart_upload()
            meta.result_status = ResultStatus.ERROR

    @abstractmethod
    def _has_expired_token(self, response: requests.Response) -> bool:
        pass

    def _send_request_with_retry(
        self,
        verb: str,
        get_request_args: Callable[[], tuple[bytes, dict[str, Any]]],
        retry_id: int,
    ) -> requests.Response:
        rest_call = METHODS[verb]
        url = b""
        conn = None
        if self.meta.sfagent and self.meta.sfagent._cursor.connection:
            conn = self.meta.sfagent._cursor.connection

        while self.retry_count[retry_id] < self.max_retry:
            cur_timestamp = self.credentials.timestamp
            url, rest_kwargs = get_request_args()
            try:
                if conn:
                    with conn._rest._use_requests_session(url) as session:
                        logger.debug(f"storage client request with session {session}")
                        response = session.request(verb, url, **rest_kwargs)
                else:
                    logger.debug("storage client request with new session")
                    response = rest_call(url, **rest_kwargs)

                if self._has_expired_presigned_url(response):
                    self._update_presigned_url()
                else:
                    self.last_err_is_presigned_url = False
                    if response.status_code in self.TRANSIENT_HTTP_ERR:
                        time.sleep(
                            min(
                                # TODO should SLEEP_UNIT come from the parent
                                #  SnowflakeConnection and be customizable by users?
                                (2 ** self.retry_count[retry_id]) * self.SLEEP_UNIT,
                                self.SLEEP_MAX,
                            )
                        )
                        self.retry_count[retry_id] += 1
                    elif self._has_expired_token(response):
                        self.credentials.update(cur_timestamp)
                    else:
                        return response
            except self.TRANSIENT_ERRORS as e:
                self.last_err_is_presigned_url = False
                time.sleep(
                    min(
                        (2 ** self.retry_count[retry_id]) * self.SLEEP_UNIT,
                        self.SLEEP_MAX,
                    )
                )
                logger.warning(f"{verb} with url {url} failed for transient error: {e}")
                self.retry_count[retry_id] += 1
        else:
            raise RequestExceedMaxRetryError(
                f"{verb} with url {url} failed for exceeding maximum retries."
            )

    def prepare_download(self) -> None:
        # TODO: add nicer error message for when target directory is not writeable
        #  but this should be done before we get here
        base_dir = os.path.dirname(self.full_dst_file_name)
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

        # HEAD
        file_header = self.get_file_header(self.meta.real_src_file_name)

        if file_header and file_header.encryption_metadata:
            self.encryption_metadata = file_header.encryption_metadata

        self.num_of_chunks = 1
        if file_header and file_header.content_length:
            self.meta.src_file_size = file_header.content_length
            if (
                self.chunked_transfer
                and self.meta.src_file_size > self.meta.multipart_threshold
            ):
                self.num_of_chunks = ceil(file_header.content_length / self.chunk_size)

        # Preallocate encrypted file.
        with self.intermediate_dst_path.open("wb+") as fd:
            fd.truncate(self.meta.src_file_size)

    def write_downloaded_chunk(self, chunk_id: int, data: bytes) -> None:
        """Writes given data to the temp location starting at chunk_id * chunk_size."""
        # TODO: should we use chunking and write content in smaller chunks?
        with self.intermediate_dst_path.open("rb+") as fd:
            fd.seek(self.chunk_size * chunk_id)
            fd.write(data)

    def finish_download(self) -> None:
        meta = self.meta
        if self.num_of_chunks != 0 and self.successful_transfers == self.num_of_chunks:
            meta.result_status = ResultStatus.DOWNLOADED
            if meta.encryption_material:
                logger.debug(f"encrypted data file={self.full_dst_file_name}")
                # For storage utils that do not have the privilege of
                # getting the metadata early, both object and metadata
                # are downloaded at once. In which case, the file meta will
                # be updated with all the metadata that we need and
                # then we can call get_file_header to get just that and also
                # preserve the idea of getting metadata in the first place.
                # One example of this is the utils that use presigned url
                # for upload/download and not the storage client library.
                if meta.presigned_url is not None:
                    file_header = self.get_file_header(meta.src_file_name)
                    self.encryption_metadata = file_header.encryption_metadata

                tmp_dst_file_name = SnowflakeEncryptionUtil.decrypt_file(
                    self.encryption_metadata,
                    meta.encryption_material,
                    str(self.intermediate_dst_path),
                    tmp_dir=self.tmp_dir,
                )
                shutil.move(tmp_dst_file_name, self.full_dst_file_name)
                self.intermediate_dst_path.unlink()
            else:
                logger.debug(f"not encrypted data file={self.full_dst_file_name}")
                shutil.move(str(self.intermediate_dst_path), self.full_dst_file_name)
            stat_info = os.stat(self.full_dst_file_name)
            meta.dst_file_size = stat_info.st_size
        else:
            # TODO: add more error details to result/meta
            if os.path.isfile(self.full_dst_file_name):
                os.unlink(self.full_dst_file_name)
            logger.exception(f"Failed to download a file: {self.full_dst_file_name}")
            meta.dst_file_size = -1
            meta.result_status = ResultStatus.ERROR

    def upload_chunk(self, chunk_id: int) -> None:
        new_stream = not bool(self.meta.src_stream or self.meta.intermediate_stream)
        fd = (
            self.meta.src_stream
            or self.meta.intermediate_stream
            or open(self.data_file, "rb")
        )
        try:
            if self.num_of_chunks == 1:
                _data = fd.read()
            else:
                fd.seek(chunk_id * self.chunk_size)
                _data = fd.read(self.chunk_size)
        finally:
            if new_stream:
                fd.close()
        logger.debug(f"Uploading chunk {chunk_id} of file {self.data_file}")
        self._upload_chunk(chunk_id, _data)
        logger.debug(f"Successfully uploaded chunk {chunk_id} of file {self.data_file}")

    @abstractmethod
    def _upload_chunk(self, chunk_id: int, chunk: bytes) -> None:
        pass

    @abstractmethod
    def download_chunk(self, chunk_id: int) -> None:
        pass

    # Override in GCS
    def _has_expired_presigned_url(self, response: requests.Response) -> bool:
        return False

    # Override in GCS
    def _update_presigned_url(self) -> None:
        return

    # Override in S3
    def _initiate_multipart_upload(self) -> None:
        return

    # Override in S3
    def _complete_multipart_upload(self) -> None:
        return

    # Override in S3
    def _abort_multipart_upload(self) -> None:
        return

    def delete_client_data(self) -> None:
        """Deletes the tmp_dir and closes the source stream belonging to this client.
        This function is idempotent."""
        if os.path.exists(self.tmp_dir):
            logger.debug(f"cleaning up tmp dir: {self.tmp_dir}")
            try:
                shutil.rmtree(self.tmp_dir)
            except OSError as ex:
                # it's ok to ignore the exception here because another thread might
                # have cleaned up the temp directory
                logger.debug(f"Failed to delete {self.tmp_dir}: {ex}")
        if self.meta.src_stream and not self.meta.src_stream.closed:
            self.meta.src_stream.close()

    def __del__(self) -> None:
        self.delete_client_data()



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import os.path
import re
import shutil
import stat
import string
import warnings
from pathlib import Path
from test.randomize import random_string
from textwrap import dedent
from typing import Callable, Dict, Union
from unittest import mock

import pytest
from pytest import raises

from snowflake.connector.compat import IS_WINDOWS

try:
    from snowflake.connector.config_manager import (
        CONFIG_MANAGER,
        ConfigManager,
        ConfigOption,
        ConfigSlice,
        ConfigSliceOptions,
    )
    from snowflake.connector.errors import (
        ConfigManagerError,
        ConfigSourceError,
        MissingConfigOptionError,
    )
    from snowflake.connector.sf_dirs import SFPlatformDirs, _resolve_platform_dirs
except ImportError:
    # olddriver tests
    pass


def tmp_files_helper(cwd: Path, to_create: files) -> None:
    for k, v in to_create.items():
        new_file = cwd / k
        if isinstance(v, str):
            new_file.touch()
            new_file.write_text(v)
        else:
            new_file.mkdir()
            tmp_files_helper(new_file, v)


files = Dict[str, Union[str, "files"]]


@pytest.fixture
def tmp_files(tmp_path: Path) -> Callable[[files], Path]:
    def create_tmp_files(to_create: files) -> Path:
        tmp_files_helper(tmp_path, to_create)
        return tmp_path

    return create_tmp_files


def test_incorrect_config_read(tmp_files):
    tmp_folder = tmp_files(
        {
            "config.toml": dedent(
                """
                [connections.defa
                """
            )
        }
    )
    config_file = tmp_folder / "config.toml"
    with raises(
        ConfigSourceError,
        match=re.escape(
            f"An unknown error happened while loading '{str(config_file)}'"
        ),
    ):
        ConfigManager(name="test", file_path=config_file).read_config()


def test_simple_config_read(tmp_files):
    tmp_folder = tmp_files(
        {
            "config.toml": dedent(
                """\
                [connections.snowflake]
                account = "snowflake"
                user = "snowball"
                password = "password"

                [settings]
                output_format = "yaml"
                """
            )
        }
    )
    config_file = tmp_folder / "config.toml"
    TEST_PARSER = ConfigManager(
        name="test",
        file_path=config_file,
    )
    from tomlkit import parse

    TEST_PARSER.add_option(
        name="connections",
        parse_str=parse,
    )
    settings_parser = ConfigManager(
        name="settings",
    )
    settings_parser.add_option(
        name="output_format",
        choices=("json", "yaml", "toml"),
    )
    TEST_PARSER.add_subparser(settings_parser)
    assert TEST_PARSER["connections"] == {
        "snowflake": {
            "account": "snowflake",
            "user": "snowball",
            "password": "password",
        }
    }
    assert TEST_PARSER["settings"]["output_format"] == "yaml"


def test_simple_config_read_sliced(tmp_files):
    """Same test_simple_config_read, but rerads part of the config from another file."""
    tmp_folder = tmp_files(
        {
            "config.toml": dedent(
                """\
                [settings]
                output_format = "json"
                """
            ),
            "connections.toml": dedent(
                """\
                [snowflake]
                account = "snowflake"
                user = "snowball"
                password = "password"
                """
            ),
        }
    )
    TEST_PARSER = ConfigManager(
        name="root_parser",
        file_path=tmp_folder / "config.toml",
        _slices=(
            ConfigSlice(
                tmp_folder / "connections.toml", ConfigSliceOptions(), "connections"
            ),
        ),
    )
    from tomlkit import parse

    TEST_PARSER.add_option(
        name="connections",
        parse_str=parse,
    )
    settings_parser = ConfigManager(
        name="settings",
    )
    settings_parser.add_option(
        name="output_format",
        choices=("json", "yaml", "toml"),
    )
    TEST_PARSER.add_subparser(settings_parser)
    assert TEST_PARSER["connections"] == {
        "snowflake": {
            "account": "snowflake",
            "user": "snowball",
            "password": "password",
        }
    }
    assert TEST_PARSER["settings"]["output_format"] == "json"


def test_missing_value(tmp_files):
    """Test that we handle a missing configuration option gracefully."""
    tmp_folder = tmp_files(
        {
            "config.toml": dedent(
                """\
                [connections.snowflake]
                account = "snowflake"
                user = "snowball"
                password = "password"
                """
            ),
        }
    )
    TEST_PARSER = ConfigManager(
        name="root_parser",
        file_path=tmp_folder / "config.toml",
    )
    TEST_PARSER.add_option(
        name="connections",
    )
    settings_parser = ConfigManager(
        name="settings",
    )
    settings_parser.add_option(
        name="output_format",
        choices=("json", "yaml", "toml"),
    )
    TEST_PARSER.add_subparser(settings_parser)
    assert TEST_PARSER["connections"] == {
        "snowflake": {
            "account": "snowflake",
            "user": "snowball",
            "password": "password",
        }
    }
    with pytest.raises(
        MissingConfigOptionError,
        match=re.escape(
            "Configuration option 'settings.output_format' is not defined anywhere, "
            "have you forgotten to set it in a configuration file, or "
            "environmental variable?"
        ),
    ):
        TEST_PARSER["settings"]["output_format"]


def test_missing_value_sliced(tmp_files):
    """Test that we handle a missing configuration option gracefully across multiple files."""
    tmp_folder = tmp_files(
        {
            "config.toml": dedent(
                """\
                [settings]
                """
            ),
            "connections.toml": dedent(
                """\
                [snowflake]
                account = "snowflake"
                user = "snowball"
                password = "password"
                """
            ),
        }
    )
    TEST_PARSER = ConfigManager(
        name="root_parser",
        file_path=tmp_folder / "config.toml",
        _slices=(
            ConfigSlice(
                tmp_folder / "connections.toml", ConfigSliceOptions(), "connections"
            ),
        ),
    )
    TEST_PARSER.add_option(
        name="connections",
    )
    settings_parser = ConfigManager(
        name="settings",
    )
    settings_parser.add_option(
        name="output_format",
        choices=("json", "yaml", "toml"),
    )
    TEST_PARSER.add_subparser(settings_parser)
    assert TEST_PARSER["connections"] == {
        "snowflake": {
            "account": "snowflake",
            "user": "snowball",
            "password": "password",
        }
    }
    with pytest.raises(
        MissingConfigOptionError,
        match=re.escape(
            "Configuration option 'settings.output_format' is not defined anywhere, "
            "have you forgotten to set it in a configuration file, or "
            "environmental variable?"
        ),
    ):
        TEST_PARSER["settings"]["output_format"]


def test_only_in_slice(tmp_files):
    tmp_folder = tmp_files(
        {
            "config.toml": dedent(
                """\
                [settings]
                [connections.snowflake]
                account = "snowflake"
                user = "snowball"
                password = "password"
                """
            ),
        }
    )
    TEST_PARSER = ConfigManager(
        name="root_parser",
        file_path=tmp_folder / "config.toml",
        _slices=(
            ConfigSlice(
                tmp_folder / "connections.toml",
                ConfigSliceOptions(
                    only_in_slice=True,
                ),
                "connections",
            ),
        ),
    )
    TEST_PARSER.add_option(
        name="connections",
    )
    settings_parser = ConfigManager(
        name="settings",
    )
    settings_parser.add_option(
        name="output_format",
        choices=("json", "yaml", "toml"),
    )
    TEST_PARSER.add_subparser(settings_parser)
    with pytest.raises(
        ConfigSourceError,
        match="Configuration option 'connections' is not defined.*",
    ):
        TEST_PARSER["connections"]


def test_simple_nesting(monkeypatch, tmp_path):
    c1 = ConfigManager(name="test", file_path=tmp_path / "config.toml")
    c2 = ConfigManager(name="sb")
    c3 = ConfigManager(name="sb")
    c3.add_option(name="b", parse_str=lambda e: e.lower() == "true")
    c2.add_subparser(c3)
    c1.add_subparser(c2)
    with monkeypatch.context() as m:
        m.setenv("SNOWFLAKE_SB_SB_B", "TrUe")
        assert c1["sb"]["sb"]["b"] is True


def test_complicated_nesting(monkeypatch, tmp_path):
    c_file = tmp_path / "config.toml"
    c1 = ConfigManager(file_path=c_file, name="root_parser")
    c2 = ConfigManager(file_path=tmp_path / "config2.toml", name="sp")
    c2.add_option(name="b", parse_str=lambda e: e.lower() == "true")
    c1.add_subparser(c2)
    c_file.write_text(
        dedent(
            """\
            [connections.default]
            user="testuser"
            account="testaccount"
            password="testpassword"

            [sp]
            b = true
            """
        )
    )
    assert c1["sp"]["b"] is True


def test_error_missing_file_path():
    with pytest.raises(
        ConfigManagerError,
        match="ConfigManager is trying to read config file," " but it doesn't have one",
    ):
        ConfigManager(name="test_parser").read_config()


def test_error_invalid_toml(tmp_path):
    c_file = tmp_path / "c.toml"
    c_file.write_text(
        dedent(
            """\
            invalid toml file
            """
        )
    )
    with pytest.raises(
        ConfigSourceError,
        match=re.escape(f"An unknown error happened while loading '{str(c_file)}'"),
    ):
        ConfigManager(
            name="test_parser",
            file_path=c_file,
        ).read_config()


def test_error_child_conflict():
    cp = ConfigManager(name="test_parser")
    cp.add_subparser(ConfigManager(name="b"))
    with pytest.raises(
        ConfigManagerError,
        match="'b' sub-manager, or option conflicts with a child element of 'test_parser'",
    ):
        cp.add_option(name="b")


def test_explicit_env_name(monkeypatch):
    rnd_string = random_string(5)
    toml_value = dedent(
        f"""\
        text = "{rnd_string}"
        """
    )
    TEST_PARSER = ConfigManager(
        name="test_parser",
    )

    from tomlkit import parse

    TEST_PARSER.add_option(name="connections", parse_str=parse, env_name="CONNECTIONS")
    with monkeypatch.context() as m:
        m.setenv("CONNECTIONS", toml_value)
        assert TEST_PARSER["connections"] == {"text": rnd_string}


def test_error_contains(monkeypatch):
    tp = ConfigManager(
        name="test_parser",
    )
    tp.add_option(name="output_format", choices=("json", "csv"))
    with monkeypatch.context() as m:
        m.setenv("SNOWFLAKE_OUTPUT_FORMAT", "toml")
        with pytest.raises(
            ConfigSourceError,
            match="The value of output_format read from environment variable "
            "is not part of",
        ):
            tp["output_format"]


def test_error_missing_item():
    tp = ConfigManager(
        name="test_parser",
    )
    with pytest.raises(
        ConfigSourceError,
        match="No ConfigManager, or ConfigOption can be found with the name 'asd'",
    ):
        tp["asd"]


def test_error_missing_fp():
    tp = ConfigManager(
        name="test_parser",
    )
    with pytest.raises(
        ConfigManagerError,
        match="ConfigManager is trying to read config file, but it doesn't have one",
    ):
        tp.read_config()


def test_missing_config_file(tmp_path):
    config_file = tmp_path / "config.toml"
    cm = ConfigManager(name="test", file_path=config_file)
    cm.add_option(name="output_format", choices=("json", "yaml"))
    with raises(
        MissingConfigOptionError,
        match="Configuration option 'output_format' is not defined anywhere.*",
    ):
        cm["output_format"]


def test_missing_config_files_sliced(tmp_path):
    config_file = tmp_path / "config.toml"
    connections_file = tmp_path / "connections.toml"
    cm = ConfigManager(
        name="test",
        file_path=config_file,
        _slices=(ConfigSlice(connections_file, ConfigSliceOptions(), "connections"),),
    )
    cm.add_option(
        name="connections",
    )
    with raises(
        MissingConfigOptionError,
        match="Configuration option 'connections' is not defined anywhere.*",
    ):
        cm["connections"]


def test_error_missing_fp_retrieve():
    tp = ConfigManager(
        name="test_parser",
    )
    tp.add_option(name="option")
    with pytest.raises(
        ConfigManagerError,
        match="Root manager 'test_parser' is missing file_path",
    ):
        tp["option"]


@pytest.mark.parametrize("version", (None, "1"))
def test_sf_dirs(tmp_path, version):
    appname = random_string(5)
    assert (
        SFPlatformDirs(
            str(tmp_path),
            appname=appname,
            appauthor=False,
            version=version,
            ensure_exists=True,
        ).user_config_path
        == tmp_path
    )


def test_config_file_resolution_sfdirs_default():
    default_loc = os.path.expanduser("~/.snowflake")
    existed_before = os.path.exists(default_loc)
    os.makedirs(default_loc, exist_ok=True)
    try:
        assert isinstance(_resolve_platform_dirs(), SFPlatformDirs)
    finally:
        if not existed_before:
            shutil.rmtree(default_loc)


def test_config_file_resolution_sfdirs_nondefault(tmp_path, monkeypatch):
    with monkeypatch.context() as m:
        m.setenv("SNOWFLAKE_HOME", str(tmp_path))
        assert isinstance(_resolve_platform_dirs(), SFPlatformDirs)


def test_config_file_resolution_non_sfdirs(tmp_path, monkeypatch):
    with monkeypatch.context() as m:
        # 2024-01-03(bwarsaw): It's not enough to remove SNOWFLAKE_HOME from the environment, because
        # _resolve_platform_dirs() defaults to ~/.snowflake, so if the user running the tests has this
        # directory, the test will fail.  Instead, ensure that this environment variable points to a
        # non-existent directory.
        fake_home = tmp_path / ".snowflake"
        m.setenv("SNOWFLAKE_HOME", str(fake_home))
        assert not isinstance(_resolve_platform_dirs(), SFPlatformDirs)


@pytest.mark.skipif(IS_WINDOWS, reason="no getuid on Windows")
def test_warn_config_file_owner(tmp_path, monkeypatch):
    c_file = tmp_path / "config.toml"
    c1 = ConfigManager(file_path=c_file, name="root_parser")
    c1.add_option(name="b", parse_str=lambda e: e.lower() == "true")
    c_file.write_text(
        dedent(
            """\
            b = true
            """
        )
    )
    c_file.chmod(stat.S_IRUSR | stat.S_IWUSR)
    with mock.patch("os.getuid", return_value=os.getuid() + 1):
        with warnings.catch_warnings(record=True) as c:
            assert c1["b"] is True
        assert len(c) == 1
        assert (
            str(c[0].message)
            == f"Bad owner or permissions on {str(c_file)}"
            + f'.\n * To change owner, run `chown $USER "{str(c_file)}"`.\n * To restrict permissions, run `chmod 0600 "{str(c_file)}"`.\n'
        )


def test_warn_config_file_permissions(tmp_path):
    c_file = tmp_path / "config.toml"
    c1 = ConfigManager(file_path=c_file, name="root_parser")
    c1.add_option(name="b", parse_str=lambda e: e.lower() == "true")
    c_file.write_text(
        dedent(
            """\
            b = true
            """
        )
    )
    c_file.chmod(stat.S_IMODE(c_file.stat().st_mode) | stat.S_IROTH)
    with warnings.catch_warnings(record=True) as c:
        assert c1["b"] is True
    assert len(c) == 1
    chmod_message = (
        f'.\n * To change owner, run `chown $USER "{str(c_file)}"`.\n * To restrict permissions, run `chmod 0600 "{str(c_file)}"`.\n'
        if not IS_WINDOWS
        else ""
    )
    assert (
        str(c[0].message)
        == f"Bad owner or permissions on {str(c_file)}" + chmod_message
    )


def test_configoption_missing_root_manager():
    with pytest.raises(
        TypeError,
        match="_root_manager cannot be None",
    ):
        ConfigOption(
            name="test_option",
            _nest_path=["test_option"],
            _root_manager=None,
        )


def test_configoption_missing_nest_path():
    with pytest.raises(
        TypeError,
        match="_nest_path cannot be None",
    ):
        ConfigOption(
            name="test_option",
            _nest_path=None,
            _root_manager=ConfigManager(name="test_manager"),
        )


def test_deprecationwarning_sub_parsers():
    with warnings.catch_warnings(record=True) as w:
        assert ConfigManager(name="test_cm")._sub_managers == {}
        assert len(w) == 0
        assert ConfigManager(name="test_cm")._sub_parsers == {}
    assert len(w) == 1
    assert issubclass(w[-1].category, DeprecationWarning)
    assert (
        str(w[-1].message)
        == "_sub_parsers has been deprecated, use _sub_managers instead"
    )


def test_deprecationwarning_add_subparser():
    with warnings.catch_warnings(record=True) as w:
        ConfigManager(name="test_cm").add_submanager(ConfigManager(name="test_cm2"))
        assert len(w) == 0
        ConfigManager(name="test_cm").add_subparser(ConfigManager(name="test_cm3"))
    assert len(w) == 1
    assert issubclass(w[-1].category, DeprecationWarning)
    assert (
        str(w[-1].message)
        == "add_subparser has been deprecated, use add_submanager instead"
    )


def test_deprecationwarning_config_parser():
    from snowflake.connector import config_manager

    with warnings.catch_warnings(record=True) as w:
        config_manager.CONFIG_MANAGER
        assert len(w) == 0
        config_manager.CONFIG_PARSER
    assert len(w) == 1
    assert issubclass(w[-1].category, DeprecationWarning)
    assert (
        str(w[-1].message)
        == "CONFIG_PARSER has been deprecated, use CONFIG_MANAGER instead"
    )
    assert config_manager.CONFIG_MANAGER is config_manager.CONFIG_PARSER


def test_configoption_default_value(tmp_path, monkeypatch):
    env_name = random_string(
        5,
        "SF_TEST_OPTION_",
        choices=string.ascii_uppercase,
    )
    conf_val = random_string(5)
    cm = ConfigManager(
        name="test_manager",
        file_path=tmp_path / "config.toml",
    )
    cm.add_option(
        name="test_option",
        env_name=env_name,
        default=conf_val,
    )
    assert cm["test_option"] == conf_val
    env_value = random_string(5)
    with monkeypatch.context() as c:
        c.setenv(env_name, env_value)
        assert cm["test_option"] == env_value


def test_defaultconnectionname(tmp_path, monkeypatch):
    c_file = tmp_path / "config.toml"
    old_path = CONFIG_MANAGER.file_path
    CONFIG_MANAGER.file_path = c_file
    CONFIG_MANAGER.conf_file_cache = None
    try:
        with monkeypatch.context() as m:
            m.delenv("SNOWFLAKE_DEFAULT_CONNECTION_NAME", raising=False)
            assert CONFIG_MANAGER["default_connection_name"] == "default"
        env_val = random_string(5, "DEF_CONN_")
        with monkeypatch.context() as m:
            m.setenv("SNOWFLAKE_DEFAULT_CONNECTION_NAME", env_val)
            assert CONFIG_MANAGER["default_connection_name"] == env_val
        assert CONFIG_MANAGER.file_path is not None
        con_name = random_string(5, "conn_")
        c_file.write_text(
            dedent(
                f"""\
                default_connection_name = "{con_name}"
                """
            )
        )
        # re-cache config file from disk
        CONFIG_MANAGER.file_path = c_file
        CONFIG_MANAGER.conf_file_cache = None
        assert CONFIG_MANAGER["default_connection_name"] == con_name
    finally:
        CONFIG_MANAGER.file_path = old_path
        CONFIG_MANAGER.conf_file_cache = None



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import time
from unittest.mock import MagicMock

try:
    from snowflake.connector.vendored.requests.exceptions import ConnectionError
except ImportError:
    from requests.exceptions import ConnectionError

try:
    from snowflake.connector.auth.by_plugin import DEFAULT_AUTH_CLASS_TIMEOUT
except ImportError:
    DEFAULT_AUTH_CLASS_TIMEOUT = 120


def zero_backoff():
    while True:
        yield 0


try:
    from snowflake.connector.connection import DEFAULT_BACKOFF_POLICY
except ImportError:
    DEFAULT_BACKOFF_POLICY = zero_backoff


def mock_connection(
    login_timeout=DEFAULT_AUTH_CLASS_TIMEOUT,
    network_timeout=None,
    socket_timeout=None,
    backoff_policy=DEFAULT_BACKOFF_POLICY,
):
    return MagicMock(
        _login_timeout=login_timeout,
        login_timeout=login_timeout,
        _network_timeout=network_timeout,
        network_timeout=network_timeout,
        _socket_timeout=socket_timeout,
        socket_timeout=socket_timeout,
        _backoff_policy=backoff_policy,
        backoff_policy=backoff_policy,
    )


def mock_request_with_action(next_action, sleep=None):
    def mock_request(*args, **kwargs):
        if sleep is not None:
            time.sleep(sleep)
        if next_action == "RETRY":
            return MagicMock(
                status_code=503,
                close=lambda: None,
            )
        elif next_action == "ERROR":
            raise ConnectionError()

    return mock_request



#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import logging
import os
from tempfile import NamedTemporaryFile

import pytest

pytestmark = pytest.mark.skipolddriver  # old test driver tests won't run this module


import snowflake.connector

try:
    from snowflake.connector.connection_diagnostic import (
        ConnectionDiagnostic,
        _decode_dict,
    )
except ImportError:
    pass


def test_https_host_report(caplog):
    connection_diag = ConnectionDiagnostic(
        account="test", host="test.snowflakecomputing.com"
    )
    https_host_report = connection_diag._ConnectionDiagnostic__https_host_report
    https_host_report(
        host="client-telemetry.snowflakecomputing.com",
        port=443,
        host_type="OUT_OF_BAND_TELEMETRY",
    )

    assert "OUT_OF_BAND_TELEMETRY: client-telemetry.snowflakecomputing.com" in ".".join(
        connection_diag.test_results["OUT_OF_BAND_TELEMETRY"]
    )


def test_test_socket_get_cert(caplog):
    connection_diag = ConnectionDiagnostic(
        account="test", host="test.snowflakecomputing.com"
    )
    test_socket_get_cert = connection_diag._ConnectionDiagnostic__test_socket_get_cert
    result = test_socket_get_cert(
        host="client-telemetry.snowflakecomputing.com",
        port=443,
        host_type="OUT_OF_BAND_TELEMETRY",
    )

    assert "BEGIN CERTIFICATE" in result


def test_decode_dict():
    test_dict = {b"CN": b"client-telemetry.snowflakecomputing.com"}
    result = _decode_dict(test_dict)

    assert result == {
        "CN": "client-telemetry.snowflakecomputing.com"
    }, "_decode_dict method failed, binary dict not converted."


def test_exception_raise_during_diag_fail(monkeypatch, caplog):
    def mock_run_post_test(self):
        raise ValueError("Diagnostic Test Failure")

    monkeypatch.setattr(
        snowflake.connector.connection_diagnostic.ConnectionDiagnostic,
        "run_post_test",
        mock_run_post_test,
    )

    try:
        snowflake.connector.connect(
            account="testaccount",
            user="testuser",
            password="testpassword",
            database="TESTDB",
            warehouse="TESTWH",
            enable_connection_diag=True,
        )
    except Exception:
        pass

    assert "Diagnostic Test Failure" in caplog.text


def test_report_message_if_bad_allowlist(caplog):
    caplog.set_level(logging.DEBUG)
    with NamedTemporaryFile("w+", suffix=".json", delete=False) as tmp_file:
        tmp_file.write(
            "This function has been deprecated. Use SYSTEM$ALLOWLIST instead."
        )
        tmp_file.flush()

    try:
        snowflake.connector.connect(
            account="testaccount",
            user="testuser",
            password="testpassword",
            database="TESTDB",
            warehouse="TESTWH",
            enable_connection_diag=True,
            connection_diag_allowlist_path=tmp_file.name,
        )
    except Exception:
        pass

    os.unlink(tmp_file.name)
    assert "Allowlist is not a valid list of json objects" in caplog.text


def test_validate_allowlist_file_works(caplog):
    caplog.set_level(logging.DEBUG)
    with NamedTemporaryFile("w+", suffix=".json", delete=False) as tmp_file:
        tmp_file.write('[{"host":"s3.amazonaws.com","port":443,"type":"STAGE"}]')
        tmp_file.flush()

    try:
        snowflake.connector.connect(
            account="testaccount",
            user="testuser",
            password="testpassword",
            database="TESTDB",
            warehouse="TESTWH",
            enable_connection_diag=True,
            connection_diag_allowlist_path=tmp_file.name,
        )
    except Exception:
        pass

    os.unlink(tmp_file.name)
    assert "STAGE: s3.amazonaws.com" in caplog.text


def test_validate_whitelist_file_works(caplog):
    caplog.set_level(logging.DEBUG)
    with NamedTemporaryFile("w+", suffix=".json", delete=False) as tmp_file:
        tmp_file.write('[{"host":"s3.amazonaws.com","port":443,"type":"STAGE"}]')
        tmp_file.flush()

    try:
        snowflake.connector.connect(
            account="testaccount",
            user="testuser",
            password="testpassword",
            database="TESTDB",
            warehouse="TESTWH",
            enable_connection_diag=True,
            connection_diag_whitelist_path=tmp_file.name,
        )
    except Exception:
        pass

    os.unlink(tmp_file.name)
    assert "STAGE: s3.amazonaws.com" in caplog.text



#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

from unittest.mock import Mock

import snowflake.connector.telemetry
from snowflake.connector.description import CLIENT_NAME, SNOWFLAKE_CONNECTOR_VERSION


def test_telemetry_data_to_dict():
    """Tests that TelemetryData instances are properly converted to dicts."""
    assert snowflake.connector.telemetry.TelemetryData({}, 2000).to_dict() == {
        "message": {},
        "timestamp": "2000",
    }

    d = {"type": "test", "query_id": "1", "value": 20}
    assert snowflake.connector.telemetry.TelemetryData(d, 1234).to_dict() == {
        "message": d,
        "timestamp": "1234",
    }


def get_client_and_mock():
    rest_call = Mock()
    rest_call.return_value = {"success": True}
    rest = Mock()
    rest.attach_mock(rest_call, "request")
    client = snowflake.connector.telemetry.TelemetryClient(rest, 2)
    return client, rest_call


def test_telemetry_simple_flush():
    """Tests that metrics are properly enqueued and sent to telemetry."""
    client, rest_call = get_client_and_mock()

    client.add_log_to_batch(snowflake.connector.telemetry.TelemetryData({}, 2000))
    assert rest_call.call_count == 0

    client.add_log_to_batch(snowflake.connector.telemetry.TelemetryData({}, 3000))
    assert rest_call.call_count == 1


def test_telemetry_close():
    """Tests that remaining metrics are flushed on close."""
    client, rest_call = get_client_and_mock()

    client.add_log_to_batch(snowflake.connector.telemetry.TelemetryData({}, 2000))
    assert rest_call.call_count == 0

    client.close()
    assert rest_call.call_count == 1
    assert client.is_closed


def test_telemetry_close_empty():
    """Tests that no calls are made on close if there are no metrics to flush."""
    client, rest_call = get_client_and_mock()

    client.close()
    assert rest_call.call_count == 0
    assert client.is_closed


def test_telemetry_send_batch():
    """Tests that metrics are sent with the send_batch method."""
    client, rest_call = get_client_and_mock()

    client.add_log_to_batch(snowflake.connector.telemetry.TelemetryData({}, 2000))
    assert rest_call.call_count == 0

    client.send_batch()
    assert rest_call.call_count == 1


def test_telemetry_send_batch_empty():
    """Tests that send_batch does nothing when there are no metrics to send."""
    client, rest_call = get_client_and_mock()

    client.send_batch()
    assert rest_call.call_count == 0


def test_telemetry_send_batch_clear():
    """Tests that send_batch clears the first batch and will not send anything on a second call."""
    client, rest_call = get_client_and_mock()

    client.add_log_to_batch(snowflake.connector.telemetry.TelemetryData({}, 2000))
    assert rest_call.call_count == 0

    client.send_batch()
    assert rest_call.call_count == 1

    client.send_batch()
    assert rest_call.call_count == 1


def test_telemetry_auto_disable():
    """Tests that the client will automatically disable itself if a request fails."""
    client, rest_call = get_client_and_mock()
    rest_call.return_value = {"success": False}

    client.add_log_to_batch(snowflake.connector.telemetry.TelemetryData({}, 2000))
    assert client.is_enabled()

    client.send_batch()
    assert not client.is_enabled()


def test_telemetry_add_batch_disabled():
    """Tests that the client will not add logs if disabled."""
    client, _ = get_client_and_mock()

    client.disable()
    client.add_log_to_batch(snowflake.connector.telemetry.TelemetryData({}, 2000))

    assert client.buffer_size() == 0


def test_telemetry_send_batch_disabled():
    """Tests that the client will not send logs if disabled."""
    client, rest_call = get_client_and_mock()

    client.add_log_to_batch(snowflake.connector.telemetry.TelemetryData({}, 2000))
    assert client.buffer_size() == 1

    client.disable()

    client.send_batch()
    assert client.buffer_size() == 1
    assert rest_call.call_count == 0


def test_generate_telemetry_data_dict_with_basic_info():
    assert snowflake.connector.telemetry.generate_telemetry_data_dict() == {
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: CLIENT_NAME,
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: SNOWFLAKE_CONNECTOR_VERSION,
        snowflake.connector.telemetry.TelemetryField.KEY_SOURCE.value: CLIENT_NAME,
    }

    assert snowflake.connector.telemetry.generate_telemetry_data_dict(from_dict={}) == {
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: CLIENT_NAME,
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: SNOWFLAKE_CONNECTOR_VERSION,
        snowflake.connector.telemetry.TelemetryField.KEY_SOURCE.value: CLIENT_NAME,
    }

    assert snowflake.connector.telemetry.generate_telemetry_data_dict(
        from_dict={"key": "value"}
    ) == {
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: CLIENT_NAME,
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: SNOWFLAKE_CONNECTOR_VERSION,
        snowflake.connector.telemetry.TelemetryField.KEY_SOURCE.value: CLIENT_NAME,
        "key": "value",
    }

    assert snowflake.connector.telemetry.generate_telemetry_data_dict(
        from_dict={
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: "CUSTOM_CLIENT_NAME",
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: "1.2.3",
            "key": "value",
        }
    ) == {
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: "CUSTOM_CLIENT_NAME",
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: "1.2.3",
        snowflake.connector.telemetry.TelemetryField.KEY_SOURCE.value: CLIENT_NAME,
        "key": "value",
    }

    mock_connection = Mock()
    mock_connection.application = "test_application"
    assert snowflake.connector.telemetry.generate_telemetry_data_dict(
        from_dict={
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: "CUSTOM_CLIENT_NAME",
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: "1.2.3",
            "key": "value",
        },
        connection=mock_connection,
    ) == {
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: "CUSTOM_CLIENT_NAME",
        snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: "1.2.3",
        snowflake.connector.telemetry.TelemetryField.KEY_SOURCE.value: mock_connection.application,
        "key": "value",
    }


def test_generate_telemetry_data():
    telemetry_data = (
        snowflake.connector.telemetry.TelemetryData.from_telemetry_data_dict(
            from_dict={}, timestamp=123
        )
    )
    assert (
        telemetry_data.message
        == {
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: CLIENT_NAME,
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: SNOWFLAKE_CONNECTOR_VERSION,
            snowflake.connector.telemetry.TelemetryField.KEY_SOURCE.value: CLIENT_NAME,
        }
        and telemetry_data.timestamp == 123
    )

    mock_connection = Mock()
    mock_connection.application = "test_application"
    telemetry_data = (
        snowflake.connector.telemetry.TelemetryData.from_telemetry_data_dict(
            from_dict={},
            timestamp=123,
            connection=mock_connection,
        )
    )
    assert (
        telemetry_data.message
        == {
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: CLIENT_NAME,
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: SNOWFLAKE_CONNECTOR_VERSION,
            snowflake.connector.telemetry.TelemetryField.KEY_SOURCE.value: mock_connection.application,
        }
        and telemetry_data.timestamp == 123
    )

    telemetry_data = (
        snowflake.connector.telemetry.TelemetryData.from_telemetry_data_dict(
            from_dict={"key": "value"},
            timestamp=123,
            connection=mock_connection,
        )
    )
    assert (
        telemetry_data.message
        == {
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_TYPE.value: CLIENT_NAME,
            snowflake.connector.telemetry.TelemetryField.KEY_DRIVER_VERSION.value: SNOWFLAKE_CONNECTOR_VERSION,
            snowflake.connector.telemetry.TelemetryField.KEY_SOURCE.value: mock_connection.application,
            "key": "value",
        }
        and telemetry_data.timestamp == 123
    )



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

import datetime
import logging
import os
import os.path
import pickle
import stat
import time
from unittest import mock

import pytest

from snowflake.connector.compat import IS_WINDOWS

try:
    import snowflake.connector.cache as cache
except ImportError:
    cache = None

# Used to insert entries that expire instantaneously
NO_LIFETIME = datetime.timedelta(seconds=0)


class BrokenReadingPickleObject:
    def __init__(self):
        self.val = "string"

    def __getstate__(self):
        return {"val": "string"}

    def __setstate__(self, state):
        raise ModuleNotFoundError()


class BrokenWritingPickleObject:
    def __init__(self):
        self.val = "string"

    def __getstate__(self):
        raise ModuleNotFoundError()

    def __setstate__(self, state):
        self.val = "string"


class TestSFDictCache:
    def test_simple_usage(self):
        c = cache.SFDictCache.from_dict({1: "a", 2: "b"})
        assert 1 in c and 2 in c
        assert c[1] == "a"
        assert c[2] == "b"

    def test_miss(self):
        c = cache.SFDictCache.from_dict(
            {"a": 1},
        )
        with pytest.raises(KeyError):
            c["b"]

    def test_expiration(self):
        c = cache.SFDictCache.from_dict(
            {"a": 1},
            entry_lifetime=0,
        )
        with pytest.raises(KeyError):
            c[1]

    def test_access_empty(self):
        c = cache.SFDictCache()
        with pytest.raises(KeyError):
            c[1]

    # The rest of tests test that SFDictCache acts like a regular dictionary.

    def test_cast_list(self):
        c = cache.SFDictCache.from_dict({"a": 1, "b": 2})
        assert list(c) == ["a", "b"]

    def test_access(self):
        c = cache.SFDictCache.from_dict({"a": 1})
        assert c["a"] == 1
        c._entry_lifetime = NO_LIFETIME
        c["b"] = 2
        with pytest.raises(KeyError):
            c["b"]

    def test_delete(self):
        c = cache.SFDictCache.from_dict({"a": 1})
        del c["a"]
        assert len(c.keys()) == 0

    def test_contains(self):
        c = cache.SFDictCache.from_dict({"a": 1})
        assert "a" in c
        c._entry_lifetime = NO_LIFETIME
        c["b"] = 2
        assert "b" not in c
        assert "c" not in c

    def test_iter(self):
        c = cache.SFDictCache.from_dict({1: "a", 2: "b"})
        counter = 1
        # Make sure that this filters out expired entries
        c._entry_lifetime = NO_LIFETIME
        c["a"] = 100
        for e in iter(c):
            assert e == counter
            # Make sure that cache can be modified while iterating
            del c[e]
            counter += 1
        assert len(c._cache) == 0

    def test_clear(self):
        c = cache.SFDictCache.from_dict({1: "a", 2: "b"})
        assert len(c.keys()) == 2
        c.clear()
        assert len(c.keys()) == 0

    def test_get(self):
        c = cache.SFDictCache.from_dict({1: "a", 2: "b"})
        assert c.get("a", -999) == -999
        assert c.get(1) == "a"
        # Make sure that this filters out expired entries
        c._entry_lifetime = NO_LIFETIME
        c["d"] = 4
        assert c.get("d") is None
        with c._lock:
            assert c._getitem_non_locking(1) == "a"

    def test_items(self):
        c = cache.SFDictCache()
        assert c.items() == []
        c["a"] = 1
        c["b"] = 2
        # Make sure that this filters out expired entries
        c._entry_lifetime = NO_LIFETIME
        c["c"] = 3
        assert c.items() == [("a", 1), ("b", 2)]

    def test_keys(self):
        c = cache.SFDictCache()
        assert list(c.keys()) == []
        c["a"] = 1
        c["b"] = 2
        # Make sure that this filters out expired entries
        c._entry_lifetime = NO_LIFETIME
        c["c"] = 3
        assert c.keys() == ["a", "b"]

    def test_update(self):
        c = cache.SFDictCache()
        c.update({"a": 1})
        other = cache.SFDictCache.from_dict({"b": 2, "c": 3})
        c["b"] = 4
        c.update(other)
        assert c.items() == [("a", 1), ("b", 2), ("c", 3)]
        # Make sure that this filters out expired entries
        c._entry_lifetime = NO_LIFETIME
        c["d"] = 4
        assert c.items() == [("a", 1), ("b", 2), ("c", 3)]

    def test_update_newer(self):
        c = cache.SFDictCache()
        c.update({"a": 1})
        with mock.patch(
            "snowflake.connector.cache.now",
            lambda: datetime.datetime.now() + datetime.timedelta(seconds=1),
        ):
            other = cache.SFDictCache.from_dict({"a": 2, "b": 4, "c": 2})
        with mock.patch(
            "snowflake.connector.cache.now",
            lambda: datetime.datetime.now() + datetime.timedelta(seconds=2),
        ):
            c["b"] = 2
        c.update_newer(other)
        assert c.items() == [("a", 2), ("b", 2), ("c", 2)]
        # Make sure that this filters out expired entries
        c._entry_lifetime = NO_LIFETIME
        c["d"] = 4
        assert c.items() == [("a", 2), ("b", 2), ("c", 2)]

    def test_values(self):
        c = cache.SFDictCache()
        assert c.values() == []
        c["a"] = 1
        c["b"] = 2
        # Make sure that this filters out expired entries
        c._entry_lifetime = NO_LIFETIME
        c["c"] = 3
        assert c.values() == [1, 2]

    def test_telemetry(self):
        c = cache.SFDictCache.from_dict({"a": 1, "b": 2})
        assert c.telemetry == {
            "hit": 0,
            "miss": 0,
            "expiration": 0,
            "size": 2,
        }
        c["a"] = 1
        assert c.telemetry["hit"] == 0
        assert c["a"] == 1
        assert c.telemetry["hit"] == 1
        with pytest.raises(KeyError):
            c["c"]
        assert c.telemetry["miss"] == 1
        # Make sure that this filters out expired entries
        c._entry_lifetime = NO_LIFETIME
        c["c"] = 3
        with pytest.raises(KeyError):
            c["c"]
        assert c.telemetry["expiration"] == 1
        assert c.get("c") is None
        assert c.telemetry["miss"] == 2
        # These functions should not affect any numbers other than expirations
        c["c"] = 3  # expired
        assert c.values() == [1, 2]
        assert c.keys() == ["a", "b"]
        assert c.items() == [("a", 1), ("b", 2)]
        assert c.telemetry == {
            "hit": 1,
            "miss": 2,
            "expiration": 2,
            "size": 2,
        }
        assert "b" in c
        c.clear()
        assert c.telemetry == {
            "hit": 0,
            "miss": 0,
            "expiration": 0,
            "size": 0,
        }
        c.update({"a": 1, "b": 2})
        assert c.telemetry == {
            "hit": 0,
            "miss": 0,
            "expiration": 0,
            "size": 2,
        }


if cache is not None:

    class AlwaysSaveSFDictFileCache(cache.SFDictFileCache):
        def _should_save(self):
            return True

    class NeverSaveSFDictFileCache(cache.SFDictFileCache):
        def _should_save(self):
            return False


class TestSFDictFileCache:
    # Since the __getitem__ is overwritten copied over tests to test it
    def test_simple_usage(self, tmpdir):
        c = cache.SFDictFileCache.from_dict(
            {1: "a", 2: "b"},
            file_path=os.path.join(tmpdir, "c.txt"),
        )
        assert 1 in c and 2 in c
        assert c[1] == "a"
        assert c[2] == "b"
        assert c.telemetry == {
            "hit": 4,
            "miss": 0,
            "expiration": 0,
            "size": 2,
        }

    def test_miss(self, tmpdir):
        c = cache.SFDictFileCache.from_dict(
            {"a": 1},
            file_path=os.path.join(tmpdir, "c.txt"),
        )
        with pytest.raises(KeyError):
            c["b"]
        assert c.telemetry == {
            "hit": 0,
            "miss": 1,
            "expiration": 0,
            "size": 1,
        }

    def test_expiration(self, tmpdir):
        tmp_file = os.path.join(tmpdir, "c.txt")
        c = NeverSaveSFDictFileCache.from_dict(
            {"a": 1},
            entry_lifetime=0,
            file_path=tmp_file,
        )
        with pytest.raises(KeyError):
            c["a"]
        assert c.telemetry == {
            "hit": 0,
            "miss": 0,
            "expiration": 1,
            "size": 0,
        }
        c2 = AlwaysSaveSFDictFileCache(file_path=tmp_file)
        c["a"] = 1
        c2["a"] = 2
        assert c2["a"] == 2
        assert c["a"] == 2

    # The rest of the tests test the file cache portion

    def test_simple_miss(self, tmpdir):
        """Check whether miss triggers a load."""
        c = AlwaysSaveSFDictFileCache.from_dict(
            {"a": 1, "b": 2}, file_path=os.path.join(tmpdir, "c.txt")
        )
        c2 = pickle.loads(pickle.dumps(c))
        c2["c"] = 3
        with mock.patch(
            "snowflake.connector.cache.getmtime",
            lambda path: os.path.getmtime(path) + 1,
        ):
            assert c["c"] == 3

    def test_simple_expiration(self, tmpdir):
        """Check whether expiration triggers a load."""
        c = AlwaysSaveSFDictFileCache.from_dict(
            {"a": 1, "b": 2}, file_path=os.path.join(tmpdir, "c.txt")
        )
        c2 = pickle.loads(pickle.dumps(c))
        c2._entry_lifetime = NO_LIFETIME
        c2["c"] = 3
        c["c"] = 3
        with mock.patch(
            "snowflake.connector.cache.getmtime",
            lambda path: os.path.getmtime(path) + 1,
        ):
            assert c2["c"] == 3

    def test_filelock_hang(self, tmpdir, caplog):
        """Check whether if file lock is held save doesn't hang."""
        c = AlwaysSaveSFDictFileCache.from_dict(
            {"a": 1, "b": 2}, file_path=os.path.join(tmpdir, "c.txt")
        )
        c2 = pickle.loads(pickle.dumps(c))
        with c2._file_lock:
            with caplog.at_level(logging.DEBUG, logger="snowflake.connector.cache"):
                # c will dump cache to file every time an item is set as it's AlwaysSaveSFDictFileCache
                # so the program exeucted at this point, c._cache_modified is false as it has the latest cache
                # changes, we set force_flush to True
                with c._lock:
                    assert not c._save(force_flush=True)
            assert (
                "snowflake.connector.cache",
                logging.DEBUG,
                f"acquiring {c._file_lock_path} timed out, skipping saving...",
            ) in caplog.record_tuples

    def test_pickle(self, tmpdir):
        c = AlwaysSaveSFDictFileCache(file_path=os.path.join(tmpdir, "cache.txt"))
        c2 = pickle.loads(pickle.dumps(c))
        assert c is not c2
        assert c._lock is not c2._lock
        assert c._file_lock is not c2._file_lock

        c.arribute_shoud_not_be_pickled = "value"
        c2 = pickle.loads(pickle.dumps(c))
        assert c.arribute_shoud_not_be_pickled
        assert c is not c2
        assert c._lock is not c2._lock
        assert c._file_lock is not c2._file_lock
        assert not hasattr(c2, "arribute_shoud_not_be_pickled")
        for attr in cache.SFDictFileCache._ATTRIBUTES_TO_PICKLE:
            assert hasattr(c2, attr) and getattr(c2, attr) == getattr(c, attr)

    def test_precise_save_load(self, tmpdir):
        c1 = NeverSaveSFDictFileCache(file_path=os.path.join(tmpdir, "cache.txt"))
        c2 = pickle.loads(pickle.dumps(c1))
        c1["a"] = 1
        c1["b"] = 1
        assert c1.save()
        assert c2.telemetry == {
            "hit": 0,
            "miss": 0,
            "expiration": 0,
            "size": 0,
        }
        assert c2.load()
        assert c2.telemetry == {
            "hit": 0,
            "miss": 0,
            "expiration": 0,
            "size": 2,
        }

    def test_clear_expired_entries(self, tmpdir):
        c1 = NeverSaveSFDictFileCache(file_path=os.path.join(tmpdir, "cache.txt"))
        c2 = pickle.loads(pickle.dumps(c1))
        c1["a"] = 1
        c1._entry_lifetime = NO_LIFETIME
        c1["b"] = 1
        c1["c"] = 1
        assert c1.save()  # Save calls self._clear_expired_entries()
        assert c2.load()
        assert c2.keys() == ["a"]

    def test_path_choosing(self, tmpdir):
        tmp_file = os.path.join(tmpdir, "cache.txt")
        non_existent_file = "/nowhere/cache.txt"
        platforms = ("linux", "darwin", "windows")
        wrong_paths = {p: non_existent_file for p in platforms}
        for platform in platforms:
            with mock.patch(
                "snowflake.connector.cache.platform.system", return_value=platform
            ):
                assert (
                    cache.SFDictFileCache(
                        file_path={**wrong_paths, platform: tmp_file},
                    ).file_path
                    == tmp_file
                )
        # Test fallback too
        with mock.patch(
            "snowflake.connector.cache.platform.system",
            return_value="BSD",
        ):
            assert (
                cache.SFDictFileCache(
                    file_path={
                        "linux": tmp_file,
                        "windows": non_existent_file,
                    },
                ).file_path
                == tmp_file
            )

    def test_read_only(self, tmpdir):
        if IS_WINDOWS:
            pytest.skip("chmod does not work on Windows")
        os.chmod(tmpdir, stat.S_IRUSR | stat.S_IXUSR)
        with pytest.raises(
            PermissionError,
            match=f"Cache folder is not writeable: '{tmpdir}'",
        ):
            cache.SFDictFileCache(file_path=os.path.join(tmpdir, "cache.txt"))

    def test_clear(self, tmpdir):
        cache_path = os.path.join(tmpdir, "cache.txt")
        c = AlwaysSaveSFDictFileCache(file_path=cache_path)
        c.clear()
        assert not c._cache
        assert os.path.exists(cache_path)
        # Make sure not existing cache file doesn't raise Exception during cache
        os.unlink(cache_path)
        c["a"] = 1
        assert os.path.exists(cache_path)

    def test_load_with_expired_entries(self, tmpdir):
        # Test: https://snowflakecomputing.atlassian.net/browse/SNOW-698526

        # create cache first
        cache_path = os.path.join(tmpdir, "cache.txt")
        c1 = cache.SFDictFileCache(file_path=cache_path)
        c1["a"] = 1
        c1.save()

        # load cache
        c2 = cache.SFDictFileCache(
            file_path=cache_path, entry_lifetime=int(NO_LIFETIME.total_seconds())
        )
        c2["b"] = 2
        c2["c"] = 3
        # cache will expire immediately due to the NO_LIFETIME setting
        # when loading again, clearing cache logic will be triggered
        # load will trigger clear expired entries, and then further call _getitem
        c2.load()

        assert len(c2) == 1 and c2["a"] == 1 and c2._getitem_non_locking("a") == 1

    def test_broken_pickle_objects(self, tmpdir):
        cache_path = os.path.join(tmpdir, "cache.txt")
        c1 = NeverSaveSFDictFileCache(file_path=cache_path)
        c1["key"] = BrokenReadingPickleObject()
        assert c1.save()
        assert os.path.exists(cache_path) and os.path.isfile(cache_path)

        c2 = NeverSaveSFDictFileCache(file_path=cache_path)
        assert not c2._load()  # load should return false due to ModuleNotFound error

        cache_path = os.path.join(tmpdir, "cache2.txt")
        c1 = NeverSaveSFDictFileCache(file_path=cache_path)
        c1["key"] = BrokenWritingPickleObject()
        assert not c1.save()
        assert not os.path.exists(cache_path)

        c2 = NeverSaveSFDictFileCache(file_path=cache_path)
        assert not c2.load()  # load should return false due to no file


def test_file_is_not_updated(tmpdir):
    tmp_cache_file = os.path.join(tmpdir, "tmp_cache")
    sfcache = NeverSaveSFDictFileCache(file_path=tmp_cache_file, entry_lifetime=1)
    sfcache.update({"key": "value"})
    assert sfcache._cache_modified
    sfcache.save()  # this save call will dump cache to file because item was added
    assert not sfcache._cache_modified
    updated_time = os.path.getmtime(tmp_cache_file)
    sfcache.save()  # this save call will be a no-op since there is no update
    assert os.path.getmtime(tmp_cache_file) == updated_time
    sfcache.update({"key": "value2"})
    assert sfcache._cache_modified
    time.sleep(0.1)  # sleep 0.1 to avoid flushing too fast
    sfcache.save()  # this save call will dump cache to file
    assert not sfcache._cache_modified
    second_updated_time = os.path.getmtime(tmp_cache_file)
    assert second_updated_time > updated_time
    assert sfcache["key"] == "value2"
    assert not sfcache._cache_modified
    sfcache.save()  # this save call will be a no-op since there is no update
    assert not sfcache._cache_modified
    assert os.path.getmtime(tmp_cache_file) == second_updated_time
    time.sleep(1)
    sfcache.save()  # this save call will dump cache because cache item is expired
    assert not sfcache._cache_modified
    assert os.path.getmtime(tmp_cache_file) > second_updated_time

    cache3 = NeverSaveSFDictFileCache(file_path=tmp_cache_file, entry_lifetime=1)
    assert len(cache3) == 0
    os.unlink(tmp_cache_file)


def test_cache_do_not_write_while_set_item(tmpdir):
    tmp_cache_file = os.path.join(tmpdir, "tmp_cache")
    sfcache = NeverSaveSFDictFileCache(tmp_cache_file)
    to_update_dict = {}
    for i in range(1000):
        to_update_dict[i] = i
    sfcache.update(to_update_dict)
    # there should be no file created as setting item won't trigger creation
    assert sfcache._cache_modified
    assert not os.path.exists(tmp_cache_file)
    sfcache.save()
    assert not sfcache._cache_modified
    assert os.path.exists(tmp_cache_file)
    file_modified_time = os.path.getmtime(tmp_cache_file)
    sfcache2 = NeverSaveSFDictFileCache(tmp_cache_file)
    time.sleep(0.01)
    for i in range(1000):
        assert sfcache2[i] == i
    assert not sfcache2._cache_modified
    sfcache2.save()  # this save should be a no-op since there is no change
    assert os.path.getmtime(tmp_cache_file) == file_modified_time



#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import logging
from os import path
from unittest import mock
from unittest.mock import Mock

import pytest

from snowflake.connector import SnowflakeConnection
from snowflake.connector.constants import SHA256_DIGEST, ResultStatus

try:
    from snowflake.connector.util_text import random_string
except ImportError:
    from ..randomize import random_string

try:
    from snowflake.connector.errors import RequestExceedMaxRetryError
    from snowflake.connector.file_transfer_agent import (
        SnowflakeFileMeta,
        SnowflakeFileTransferAgent,
        StorageCredential,
    )
    from snowflake.connector.storage_client import METHODS
    from snowflake.connector.vendored.requests import Response
except ImportError:
    # Compatibility for olddriver tests
    from requests import Response

    SnowflakeFileMeta = dict
    RequestExceedMaxRetryError = None
    METHODS = {}
    megabytes = 1024 * 1024

try:  # pragma: no cover
    from snowflake.connector.gcs_storage_client import SnowflakeGCSRestClient
except ImportError:
    SnowflakeGCSRestClient = None

# We need these for our OldDriver tests. We run most up to date tests with the oldest supported driver version
try:
    from snowflake.connector.vendored import requests

    vendored_request = True
except ImportError:  # pragma: no cover
    import requests

    vendored_request = False

THIS_DIR = path.dirname(path.realpath(__file__))


@pytest.mark.parametrize("errno", [408, 429, 500, 503])
def test_upload_retry_errors(errno, tmpdir):
    """Tests whether retryable errors are handled correctly when upploading."""
    f_name = str(tmpdir.join("some_file.txt"))
    resp = requests.Response()
    resp.status_code = errno
    meta = SnowflakeFileMeta(
        name=f_name,
        src_file_name=f_name,
        stage_location_type="GCS",
        presigned_url="some_url",
        sha256_digest="asd",
    )
    if RequestExceedMaxRetryError is not None:
        mock_connection = mock.create_autospec(SnowflakeConnection)
        client = SnowflakeGCSRestClient(
            meta,
            StorageCredential({}, mock_connection, ""),
            {},
            mock_connection,
            "",
        )
    with open(f_name, "w") as f:
        f.write(random_string(15))
    if RequestExceedMaxRetryError is None:
        with mock.patch(
            "snowflake.connector.vendored.requests.put"
            if vendored_request
            else "requests.put",
            side_effect=requests.exceptions.HTTPError(response=resp),
        ):
            SnowflakeGCSUtil.upload_file(f_name, meta, None, 99, 64000)
            assert isinstance(meta.last_error, requests.exceptions.HTTPError)
            assert meta.result_status == ResultStatus.NEED_RETRY
    else:
        client.data_file = f_name

        if vendored_request:
            with mock.patch.dict(
                METHODS,
                {"PUT": lambda *a, **kw: resp},
            ):
                with pytest.raises(RequestExceedMaxRetryError):
                    # Retry quickly during unit tests
                    client.SLEEP_UNIT = 0.0
                    client.upload_chunk(0)
        else:
            # Old Driver test specific code
            with mock.patch("requests.put"):
                SnowflakeGCSUtil.upload_file(f_name, meta, None, 99, 64000)
                assert isinstance(meta.last_error, requests.exceptions.HTTPError)
                assert meta.result_status == ResultStatus.NEED_RETRY


def test_upload_uncaught_exception(tmpdir):
    """Tests whether non-retryable errors are handled correctly when uploading."""
    f_name = str(tmpdir.join("some_file.txt"))
    resp = requests.Response()
    resp.status_code = 501
    exc = requests.exceptions.HTTPError(response=resp)
    with open(f_name, "w") as f:
        f.write(random_string(15))
    agent = SnowflakeFileTransferAgent(
        mock.MagicMock(),
        f"put {f_name} @~",
        {
            "data": {
                "command": "UPLOAD",
                "src_locations": [f_name],
                "stageInfo": {
                    "locationType": "GCS",
                    "location": "",
                    "creds": {"AWS_SECRET_KEY": "", "AWS_KEY_ID": ""},
                    "region": "test",
                    "endPoint": None,
                },
                "localLocation": "/tmp",
            }
        },
    )
    with mock.patch(
        "snowflake.connector.gcs_storage_client.SnowflakeGCSRestClient.get_file_header",
    ), mock.patch(
        "snowflake.connector.gcs_storage_client.SnowflakeGCSRestClient._upload_chunk",
        side_effect=exc,
    ):
        agent.execute()
    assert agent._file_metadata[0].error_details is exc


@pytest.mark.parametrize("errno", [403, 408, 429, 500, 503])
def test_download_retry_errors(errno, tmp_path):
    """Tests whether retryable errors are handled correctly when downloading."""
    resp = requests.Response()
    resp.status_code = errno
    if errno == 403:
        pytest.skip("This behavior has changed in the move from SDKs")
    meta_info = {
        "name": "data1.txt.gz",
        "stage_location_type": "S3",
        "no_sleeping_time": True,
        "put_callback": None,
        "put_callback_output_stream": None,
        SHA256_DIGEST: "123456789abcdef",
        "dst_file_name": "data1.txt.gz",
        "src_file_name": path.join(THIS_DIR, "../data", "put_get_1.txt"),
        "overwrite": True,
    }
    meta = SnowflakeFileMeta(**meta_info)
    creds = {"AWS_SECRET_KEY": "", "AWS_KEY_ID": ""}
    cnx = mock.MagicMock(autospec=SnowflakeConnection)
    rest_client = SnowflakeGCSRestClient(
        meta,
        StorageCredential(
            creds,
            cnx,
            "GET file:/tmp/file.txt @~",
        ),
        {
            "locationType": "AWS",
            "location": "bucket/path",
            "creds": creds,
            "region": "test",
            "endPoint": None,
        },
        cnx,
        "GET file:///tmp/file.txt @~",
    )
    from snowflake.connector.storage_client import METHODS

    rest_client.SLEEP_UNIT = 0
    with mock.patch.dict(METHODS, GET=mock.MagicMock(return_value=resp)):
        with pytest.raises(
            RequestExceedMaxRetryError,
            match="GET with url .* failed for exceeding maximum retries",
        ):
            rest_client.download_chunk(0)


@pytest.mark.parametrize("errno", (501, 403))
def test_download_uncaught_exception(tmp_path, errno):
    """Tests whether non-retryable errors are handled correctly when downloading."""
    resp = requests.Response()
    resp.status_code = errno
    meta_info = {
        "name": "data1.txt.gz",
        "stage_location_type": "S3",
        "no_sleeping_time": True,
        "put_callback": None,
        "put_callback_output_stream": None,
        SHA256_DIGEST: "123456789abcdef",
        "dst_file_name": "data1.txt.gz",
        "src_file_name": path.join(THIS_DIR, "../data", "put_get_1.txt"),
        "overwrite": True,
    }
    meta = SnowflakeFileMeta(**meta_info)
    creds = {"AWS_SECRET_KEY": "", "AWS_KEY_ID": ""}
    cnx = mock.MagicMock(autospec=SnowflakeConnection)
    rest_client = SnowflakeGCSRestClient(
        meta,
        StorageCredential(
            creds,
            cnx,
            "GET file:/tmp/file.txt @~",
        ),
        {
            "locationType": "AWS",
            "location": "bucket/path",
            "creds": creds,
            "region": "test",
            "endPoint": None,
        },
        cnx,
        "GET file:///tmp/file.txt @~",
    )
    from snowflake.connector.storage_client import METHODS

    rest_client.SLEEP_UNIT = 0
    with mock.patch.dict(METHODS, GET=mock.MagicMock(return_value=resp)):
        with pytest.raises(
            requests.exceptions.HTTPError,
        ):
            rest_client.download_chunk(0)


def test_upload_put_timeout(tmp_path, caplog):
    """Tests whether timeout error is handled correctly when uploading."""
    caplog.set_level(logging.DEBUG, "snowflake.connector")
    f_name = str(tmp_path / "some_file.txt")
    resp = requests.Response()
    with open(f_name, "w") as f:
        f.write(random_string(15))
    agent = SnowflakeFileTransferAgent(
        mock.Mock(autospec=SnowflakeConnection, connection=None),
        f"put {f_name} @~",
        {
            "data": {
                "command": "UPLOAD",
                "src_locations": [f_name],
                "stageInfo": {
                    "locationType": "GCS",
                    "location": "",
                    "creds": {"AWS_SECRET_KEY": "", "AWS_KEY_ID": ""},
                    "region": "test",
                    "endPoint": None,
                },
                "localLocation": "/tmp",
            }
        },
    )
    mocked_put, mocked_head = mock.MagicMock(), mock.MagicMock()
    mocked_put.side_effect = requests.exceptions.Timeout(response=resp)
    resp = Response()
    resp.status_code = 404
    mocked_head.return_value = resp
    SnowflakeGCSRestClient.SLEEP_UNIT = 0
    from snowflake.connector.storage_client import METHODS

    with mock.patch.dict(METHODS, {"PUT": mocked_put, "HEAD": mocked_head}):
        agent.execute()
    assert (
        "snowflake.connector.storage_client",
        logging.WARNING,
        "PUT with url https://storage.googleapis.com//some_file.txt.gz failed for transient error: ",
    ) in caplog.record_tuples
    assert (
        "snowflake.connector.file_transfer_agent",
        logging.DEBUG,
        "Chunk 0 of file some_file.txt failed to transfer for unexpected exception PUT with url https://storage.googleapis.com//some_file.txt.gz failed for exceeding maximum retries.",
    ) in caplog.record_tuples


def test_download_timeout(tmp_path, caplog):
    """Tests whether timeout error is handled correctly when downloading."""
    timeout_exc = requests.exceptions.Timeout(response=requests.Response())
    meta_info = {
        "name": "data1.txt.gz",
        "stage_location_type": "S3",
        "no_sleeping_time": True,
        "put_callback": None,
        "put_callback_output_stream": None,
        SHA256_DIGEST: "123456789abcdef",
        "dst_file_name": "data1.txt.gz",
        "src_file_name": path.join(THIS_DIR, "../data", "put_get_1.txt"),
        "overwrite": True,
    }
    meta = SnowflakeFileMeta(**meta_info)
    creds = {"AWS_SECRET_KEY": "", "AWS_KEY_ID": ""}
    cnx = mock.MagicMock(autospec=SnowflakeConnection)
    rest_client = SnowflakeGCSRestClient(
        meta,
        StorageCredential(
            creds,
            cnx,
            "GET file:/tmp/file.txt @~",
        ),
        {
            "locationType": "AWS",
            "location": "bucket/path",
            "creds": creds,
            "region": "test",
            "endPoint": None,
        },
        cnx,
        "GET file:///tmp/file.txt @~",
    )
    from snowflake.connector.storage_client import METHODS

    rest_client.SLEEP_UNIT = 0
    with mock.patch.dict(METHODS, GET=mock.MagicMock(side_effect=timeout_exc)):
        exc = Exception("stop execution")
        with mock.patch.object(rest_client.credentials, "update", side_effect=exc):
            with pytest.raises(RequestExceedMaxRetryError):
                rest_client.download_chunk(0)


def test_get_file_header_none_with_presigned_url(tmp_path):
    """Tests whether default file handle created by get_file_header is as expected."""
    meta = SnowflakeFileMeta(
        name=str(tmp_path / "some_file"),
        src_file_name=str(tmp_path / "some_file"),
        stage_location_type="GCS",
        presigned_url="www.example.com",
    )
    storage_credentials = Mock()
    storage_credentials.creds = {}
    stage_info = Mock()
    connection = Mock()
    client = SnowflakeGCSRestClient(
        meta, storage_credentials, stage_info, connection, ""
    )
    file_header = client.get_file_header(meta.name)
    assert file_header is None
